/*
 *	Copyright 1995,1996 (C) by Karsten Ballder
 *				Karsten.Ballder@stud.uni-karlsruhe.de
 *
 *
 * This software comes with ABSOLUTELY NO WARRANTY; for details see
 * the section COPYRIGHT in the manual. This is free software and you
 * are welcome to redistribute it under certain conditions. 
 *
 *
 *	If you are using this program within some other software, you must
 *	state that your software is using MultiBuf, that MultiBuf is part of
 *	the KBackup package and your software must show the above copyright.
 *
 *	multibuf.c : 	program for transparently handling multi-volume
 *			archives with tar or afio
 *
 */
/*	The meaning of the three commands:
 *
 * command:	executed at end of medium (e.g. prompting, umount)
 * command2:	executed at beginning of new medium (e.g. mt rewind, mount)
 * rcommand:	executed after command2 if reading fails or wrong sequence
 *		rcommand is not followed by another command2!
 *		e.g.: umount + prompting to change medium + mount
 *
 *	I am not sure whether this is intelligent or not, maybe I should
 *	execute a command2 after each rcommand? However, like this it works,
 *	so I will keep it like this.
 */

 
#define	REVISION	"1.13"
#define	ID		"multibuf.c,v 1.13 1996/03/25 17:20:41 kbackup Exp"
#define	DATE		"1996/03/25 17:20:41"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<getopt.h>

#include	<sys/types.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<string.h>
#include	<errno.h>

#include	<assert.h>


#define	DATA_FORMAT_VERSION	1


#define	DEFAULT_BLOCKSIZE	1024
#define	MINIMUM_BLOCKSIZE	1024
#define	DEFAULT_BLOCKS		100
#define	MINIMUM_BLOCKS		5

#define DEFAULT_COMMAND "dialog --title \"KBackup's MultiBuf\" --yesno \"Continue with next volume?\" 5 40 </dev/tty >/dev/tty"
#define	DEFAULT_COMMAND2 "mt rewind -f %s</dev/tty >/dev/tty"
#define	DEFAULT_RCOMMAND "dialog --title \"KBackup's MultiBuf\" --yesno \"Detected wrong sequence number.\nRetry?\" 6 50 </dev/tty >/dev/tty"
#define MULTIBUF_HEADER		"MultiBuf by Karsten Ballder"
#define MULTIBUF_HEADER_LEN	28
#define ENDMARKER		"ENDMARKER"
#define ENDMARKER_LEN		10

#define	EXIT_WRONG_ARGS		1
#define	EXIT_CANNOT_READ	2
#define	EXIT_CANNOT_WRITE	3
#define	EXIT_WRONG_SEQUENCE	4
#define	EXIT_NO_MEM		5
#define	EXIT_FATAL_ERROR	255

size_t	bufsize 	= DEFAULT_BLOCKS*DEFAULT_BLOCKSIZE;
size_t	blocks		= DEFAULT_BLOCKS;
size_t	blocksize 	= DEFAULT_BLOCKSIZE;
size_t	bytes_written	= 0;
char *	command		= DEFAULT_COMMAND;
char *	command2	= DEFAULT_COMMAND2;
char *  rcommand	= DEFAULT_RCOMMAND;
char *	device		= "<NULL>";

char *	buffer		= NULL;
int	write_mode	= 1;
int	verbose		= 0;
int	command2enable	= 1;
int	use_header	= 0;
int	sequence_number	= 1;
int	use_endmarker	= 1;
int	file_endmarker  = 0;
int	disable_fmark_fix = 0;

char *	progname	= "multibuf";

extern char *optarg;
extern int optind, opterr, optopt;

static struct option long_opts[] =
{
	{"help",0,0,'?'},	
	{"read",0,0,'r'},
	{"write",0,0,'w'},
	{"verbose",0,0,'v'},
	{"command",1,0,'c'},
	{"command2",1,0,'C'},
	{"retry-command",1,0,'R'},
	{"begin-and-end",1,0,'2'},
	{"end-only",1,0,'1'},
	{"settings",0,0,'s'},
	{"nblocks",1,0,'n'},
	{"blocksize",1,0,'b'},
	{"exit",0,0,'x'},
	{"header",0,0,'h'},
	{"version",0,0,'V'},
	{"no-endmarker",0,0,'E'},
	{"file-endmarker",0,0,'F'},
	{"disable-filemark-fix",0,0,'D'},
	{0,0,0,0}
};
static char short_opts[]="?rwvc:sn:b:C:12xhVR:EFD";

static void print_longhelp(void)
{
	fprintf(stderr,
"MultiBuf " REVISION "\t(C) 1995 by Karsten Ballder\n"
"                           Karsten.Ballueder@stud.uni-karlsruhe.de\n"
"\n"
"USAGE: %s [options] filename\n"
"valid options:\n"
"   -r, --read		: set read mode\n"
"   -w, --write		: set write mode (default)\n"
"   -v, --verbose	: verbose progress reporting\n"
"   -b, --blocksize	: set size of each data block in bytes\n"
"   -n, --nblocks	: set number of blocks in each superblock\n"
"   -c, --command \"shell-command\"  : set command to execute at end of medium\n"
"   -C, --command2 \"shell-command\" : set command to execute at begin of medium\n"
"   -R, --retry-command \"shell-command\" : set command to execute when finding\n"
"   					   wrong sequence number\n"
"   -1, --end-only	: only execute command at end of media\n"
"   -2, --begin-and-end	: exeute commands at end and begin of media\n"
"   -s, --settings	: display settings\n"
"   -x, --exit		: exit after processing options (for testing)\n"
"   -h, --header		: add header/sequence information to data\n"
"   -E, --no-endmarker   : do not use an endmarker\n"
"   -F, --file-endmarker : write endmarker as a file on a mounted medium\n"
"   -D, --disable-filemark-fix : for debugging purpose only\n"
"   -V, --version	: print version information\n"
"\n"
"   return codes: 0=success, 1=read error, 2=write error, 3=wrong sequence\n"
, progname);
	exit(EXIT_SUCCESS);
}

void	display_settings(void)
{	
	if(write_mode)
		fprintf(stderr,"operating in write-mode\n");
	else
		fprintf(stderr,"operating in read-mode\n");

	if(use_header)
	{	fprintf(stderr,"use sequence header\n");
		if(use_endmarker)
		{	fprintf(stderr,"\tuse endmarker");
			if(file_endmarker)
				fprintf(stderr,", written as separate file\n");
			else
				fprintf(stderr,", written as tape block\n");
		}
	}	
	if(command2enable)
		fprintf(stderr,"call commands 1 and 2\n");
	else
		fprintf(stderr,"call one command only\n");

	fprintf(stderr,
		"size of one data block: %ld, number of blocks per superblock: %ld\n"
		"\t==> total buffer size: %ld\n"
		"first command=[%s]\n",
			(long)blocksize, (long)blocks,
			(long)bufsize, command);
	if(command2enable)
		fprintf(stderr,"second command=[%s]\n", command2);
	fprintf(stderr,"retry command=[%s]\n",rcommand);

}
void 	print_help(void)
{
	fprintf(stderr,"\aUSAGE: %s [options] filename\n	type %s -? for help\n", progname, progname);
}

void	process_options(int argn, char **argv)
{
	int opt_index, c;
	
	for(;;)
	{	c = getopt_long(argn,argv,short_opts,long_opts, &opt_index);
		if( c == -1)
			break;
		switch(c)
		{	
			case '?':
				print_longhelp();
				break;
			case 'r':
				write_mode = 0;
				break;
			case 'w':
				write_mode = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'c':
				command = optarg;
				break;
			case 'C':
				command2 = optarg;
				break;
			case 'R':
				rcommand = optarg;
				break;
			case 'h':
				use_header = 1;
				break;
			case '1':
				command2enable = 0;
				break;
			case '2':
				command2enable = 1;
				break;
			case 'E':
				use_endmarker = 0;
				break;
			case 'F':
				file_endmarker = 1;
				break;
			case 's':
				display_settings();
				break;
			case 'b':
				blocksize = (size_t) atol(optarg);
				if(blocksize < MINIMUM_BLOCKSIZE)
					blocksize=MINIMUM_BLOCKSIZE;
				bufsize	= blocks * blocksize;
				break;
			case 'n':
				blocks = (size_t) atol(optarg);
				if(blocks < MINIMUM_BLOCKS)
					blocks=MINIMUM_BLOCKS;
				bufsize	= blocks * blocksize;
				break;
			case 'V':
				fprintf(stderr,"MultiBuf Revision " REVISION "\n\t"
					ID "\n\t" DATE "\n");
				break;
			case 'x':
				exit(EXIT_SUCCESS);
			case 'D':
				disable_fmark_fix = 1;
				break;
			default:
				print_help();
				exit(EXIT_WRONG_ARGS);
				;
		}
	}
}



void write_loop(const char *filename)
{
	int 	output = -1, rc;
	int	nread, written;
	int	result;
	char	*buffer2, strbuf[1024];

	if(verbose)
		fprintf(stderr,"write_loop:\n");

	assert(output==-1);
	output=open(filename,O_WRONLY|O_CREAT, 0666);
	if(output < 0)
	{	fprintf(stderr,"\nmultibuf: Cannot open output file %s!\a\n", filename);
		exit(EXIT_CANNOT_WRITE);
	}

	if(use_header)
	{	strcpy(buffer,MULTIBUF_HEADER);
		sprintf(buffer+strlen(MULTIBUF_HEADER)+1,"%05d %03d %010u %010u", sequence_number++,
		DATA_FORMAT_VERSION,(unsigned) blocksize, (unsigned) blocks);
		result=write(output,buffer,blocksize);
		if(result < blocksize)
		{	fprintf(stderr,"\nmultibuf: Cannot write header to output!\a\n");
			exit(EXIT_CANNOT_WRITE);
		}
	}
	
	for(;;)
	{	nread = fread(buffer, blocksize, blocks, stdin);
		if(nread < blocks)
		{	if(nread == 0)	/* end of input data */
				break;
			if(verbose)
				fprintf(stderr,"\nOnly read %d blocks.\n", nread);
		}
		do
		{	if(nread == blocks)
				written = write(output, buffer, bufsize);
			else
			{	written = write(output, buffer, nread*blocksize);
				break;
			}
			if(written < bufsize)
			{	close(output); output=-1;
				rc = system(command);
				if((rc == 127 && errno != 0)||rc != 0)
					break;
				if(command2enable)
				{	sprintf(strbuf,command2,device);
					rc = system(strbuf);
					if((rc == 127 && errno != 0)||rc != 0)
					{	fprintf(stderr,"\nmultibuf: Cannot execute \"%s\"!\a\n", strbuf);
						break;
					}
				}
				assert(output==-1);
				output=open(filename,O_CREAT|O_WRONLY,0666);
				if(output < 0)
				{	fprintf(stderr,"\nmultibuf: Cannot open output file %s!\a\n", filename);
					exit(EXIT_CANNOT_WRITE);
				}
				if(use_header)
				{	buffer2=malloc(blocksize);
					sprintf(buffer2,"%05d", sequence_number++);
					result=write(output,buffer2,blocksize);
					if(result < blocksize)
					{	fprintf(stderr,"\nmultibuf: Cannot write sequence number!\a\n");
						exit(EXIT_CANNOT_WRITE);
					}
					free(buffer2);
				}
			}
		}while(written < bufsize);
		bytes_written += written;
		if(verbose)
			fprintf(stderr,"\r wrote %ld bytes", (long int) bytes_written);
	} /* for */
	if(verbose)
		fprintf(stderr,"\nTotal bytes written: %ld\n", (long int) bytes_written);
	close(output); output=-1;

	/* 	write an END marker after last block
		just a short file, so after EOF another read returns
	*/
	if(use_header && use_endmarker)	/* write END marker */
	{	if(file_endmarker)
		{	char *endfile = malloc(strlen(filename)+5);
			if(! endfile)
				exit(EXIT_NO_MEM);
			strcpy(endfile,filename); strcat(endfile,".END");
			assert(output==-1);
			output = open(endfile,O_CREAT|O_WRONLY,0666);
			free(endfile);
		}
		else
		{	assert(output==-1);
			output=open(filename,O_CREAT|O_WRONLY,0666);
		}
		if(output < 0)
		{	fprintf(stderr,"\nmultibuf: Cannot write end marker (%s)!\a\n",
				strerror(errno));
			exit(EXIT_CANNOT_WRITE);
		}
		strcpy(buffer,ENDMARKER);
		write(output,buffer,blocksize);
		if(verbose)
			fprintf(stderr,"\nwrote end marker\n");
		close(output); output=-1;
	}
}


void read_loop(const char *filename)
{
	int		input = -1,
			nread, written,
			result,rc,
			seq_no = -1,		/* seq_no: from file */
			retry;
	char		*buffer2, strbuf[1024];
	int		endmarker_detected=0;

	int		data_format_version=-1;
	unsigned	ublocks=(unsigned) blocks, 
			ublocksize= (unsigned) blocksize;

	if(verbose)
		fprintf(stderr,"read_loop:\n");

	retry = ! disable_fmark_fix;
	restart_read:
	assert(input == -1);
	input=open(filename,O_RDONLY);
	if(input < 0)
	{	fprintf(stderr,"multibuf: Cannot open input file %s!\a\n", filename);
		exit(EXIT_CANNOT_READ);
	}
	if(use_header)
	{	result=read(input,buffer,blocksize);
		if(result < blocksize || 
			strcmp(buffer,MULTIBUF_HEADER) != 0)
		{	close(input); input=-1;
			if(result == 0 && retry)
			{	retry = 0;
				goto restart_read;
			}
			fprintf(stderr,"multibuf: Cannot read header!\a\n");
			rc = system(rcommand); /* retry */
			if((rc == 127 && errno != 0)||rc != 0)
				goto restart_read;
			exit(EXIT_CANNOT_READ);
		}
		sscanf(buffer+strlen(MULTIBUF_HEADER)+1,"%05d %03d %010u %010u", &seq_no,
		&data_format_version, &ublocksize, &ublocks);
		if(sequence_number != seq_no)
		{	close(input); input=-1;
			rc = system(rcommand); /* retry */
			if((rc == 127 && errno != 0)||rc != 0)
				goto restart_read;
			fprintf(stderr,"\nmultibuf: Wrong sequence number!\a\n"
			"\tExpected %d, read %d\n", sequence_number, seq_no);
			exit(EXIT_WRONG_SEQUENCE);
		}
		if(data_format_version >= DATA_FORMAT_VERSION)
		{	blocks=(size_t) ublocks;
			blocksize=(size_t) ublocksize;
			if(verbose)
				fprintf(stderr,"\nset blocksize=%u, nblocks=%u\n",
					ublocksize, ublocks);
		}
		sequence_number++;
	}
	
	for(;;)
	{	do
		{	nread = read(input, buffer, bufsize);
		
			if(nread < bufsize)
			{	close(input); input=-1;
				if(use_header && use_endmarker)	/* look for end marker */
				{	endmarker_detected=0;
					if(file_endmarker)
					{	char *endfile = malloc(strlen(filename)+5);
						if(! endfile)
							exit(EXIT_NO_MEM);
						strcpy(endfile,filename); strcat(endfile,".END");
						assert(input == -1);
						input=open(endfile,O_RDONLY);
						if(input >= 0)
						{	endmarker_detected=1;
							close(input); input=-1;
						}
						free(endfile);
					}
					else /* look for normal endmarker: */
					{	retry = ! disable_fmark_fix;
					retry_endmarker:
						assert(input == -1);
						input=open(filename,O_RDONLY);
						if(input < 0)
						{
						fprintf(stderr,"\nmultiuf: Cannot open input file %s!\a\n", filename);
							exit(EXIT_CANNOT_READ);
						}
						endmarker_detected=0;
						buffer2=malloc(blocksize);
						if(buffer2 && read(input,buffer2,blocksize))
						{	if(strncmp(buffer2,ENDMARKER,ENDMARKER_LEN)==0)
							{	if(verbose)
									fprintf(stderr,"\nEND marker detected.\n");
								free(buffer2);
								endmarker_detected=1;
							}
						}
						else
						if(buffer2 && retry ) /* retry */
						{	close(input); input=-1;
							retry=0;
							if(verbose)
								fprintf(stderr,"retrying read_endmarker\n");
							goto retry_endmarker;
						}
						if(buffer2) free(buffer2);
						close(input); input=-1;
					}
				}
				if(endmarker_detected ||
					( ( ( rc = system(command) ) == 127 &&
					errno != 0) || rc != 0) )
				{	written = fwrite(buffer, nread,1, stdout);
					bytes_written += written;
					if(verbose)
						fprintf(stderr,"\r read %ld bytes", (long int) bytes_written);
					exit(EXIT_SUCCESS);
				}
				/* continue reading */
				if(command2enable)
				{	sprintf(strbuf,command2,device);
					rc = system(strbuf);
					if((rc == 127 && errno != 0)||rc != 0)
					{	fprintf(stderr,"\nmultibuf: Cannot execute \"%s\"!\a\n", strbuf);
						exit(EXIT_CANNOT_READ);
					}
				}
			restart_read2:
				retry = ! disable_fmark_fix;
			retry_read2:
				assert(input == -1);
				input=open(filename,O_RDONLY);
				if(input < 0)
				{	fprintf(stderr,"\nmultibuf: Cannot open input file %s!\a\n", filename);
					exit(1);
				}
				if(use_header)
				{	buffer2 = malloc(blocksize);
					result=read(input,buffer2, blocksize);
					if(retry && result == 0)
					{	close(input); input=-1;
						if(verbose)
							fprintf(stderr,"retrying read2\n");
						goto retry_read2;
					}
					if(result < blocksize)
					{	fprintf(stderr,"\nmultibuf: Cannot read sequence number!\a\n");
						exit(EXIT_CANNOT_READ);
					}
					sscanf(buffer2,"%05d", &seq_no);
					free(buffer2);
					if(sequence_number != seq_no)
					{	close(input); input=-1;
						if(verbose)
							fprintf(stderr,"seq no mismatch, expected: %d, got: %d\n", sequence_number, seq_no);
						rc = system(rcommand); /* retry */
						if((rc == 127 && errno != 0)||rc != 0)
							goto restart_read2;
						exit(EXIT_CANNOT_READ);
					}		
					sequence_number++;
				}
			}
		}while(nread < bufsize);
		bytes_written += nread;
		if(verbose)
			fprintf(stderr,"\r read %ld bytes", (long int) bytes_written);
		written = fwrite(buffer, nread,1, stdout);
		if(written < 1)
		{	fprintf(stderr,"\nmultibuf: Only wrote %d blocks!\a\n", nread);
			exit(EXIT_CANNOT_WRITE);
		}
		if(verbose)
			fprintf(stderr,"\r read %ld bytes", (long int) bytes_written);
	}
	close(input); input=-1;
}



int	main(int argc, char **argv)
{
	progname = argv[0];


	process_options(argc, argv);

	if(argc < 2)
	{	print_help();
		return 1;
	}
	
	device=argv[argc - 1];
	
	buffer = malloc(bufsize=blocks*blocksize);
	if(! buffer)
	{	fprintf(stderr,"multibuf: Cannot allocate memory!\a\n");
		return 1;
	}

	if(write_mode)
		write_loop(device);
	else
		read_loop(device);
	free(buffer);	
	if(verbose)
		fprintf(stderr,"\nTotal bytes written: %ld\n", (long int) bytes_written);
	return 0;
}

